/*
    Copyright (C) 1999 by Seth Galbraith

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "squawk.h"
#include "sqkview.h"
#include "sqksystm.h"

#include "iengine/engine.h"
#include "iengine/sector.h"
#include "iengine/camera.h"
#include "iengine/campos.h"
#include "iengine/light.h"
#include "imesh/thing/polygon.h"
#include "imesh/object.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "iengine/material.h"

#include "cstool/csview.h"
#include "cstool/initapp.h"
#include "csutil/scanstr.h"
#include "csutil/csstring.h"
#include "csutil/cmdhelp.h"

#include "iutil/vfs.h"
#include "iutil/plugin.h"
#include "iutil/eventq.h"
#include "iutil/eventh.h"
#include "iutil/comp.h"
#include "iutil/cfgmgr.h"
#include "iutil/cmdline.h"
#include "ivideo/graph2d.h"
#include "ivideo/graph3d.h"
#include "ivideo/txtmgr.h"
#include "ivaria/conout.h"
#include "ivaria/reporter.h"
#include "ivaria/stdrep.h"
#include "imesh/sprite3d.h"
#include "imap/parser.h"
#include "iutil/event.h"
#include "iutil/objreg.h"
#include "iutil/csinput.h"
#include "iutil/virtclk.h"

#include "debug/fpu80x86.h"

//------------------------------------------------- We need the 3D engine -----

extern sqkSystem* Sys;

sqkSystem::sqkSystem ()
{
  Sys = this;
  vc = NULL;
  frameTime = 0;
  engine = NULL;
  mapfile = NULL;
  viewLayout = 1;
  viewMargin = 0;
  LevelLoader = NULL;
  myG3D = NULL;
  myG2D = NULL;
  myVFS = NULL;
  kbd = NULL;
}

sqkSystem::~sqkSystem ()
{
  SCF_DEC_REF (vc);
  SCF_DEC_REF (myG3D);
  SCF_DEC_REF (myG2D);
  SCF_DEC_REF (myVFS);
  SCF_DEC_REF (kbd);
  if (LevelLoader)
    LevelLoader->DecRef ();
  if (engine)
    engine->DecRef ();
}

void sqkSystem::AddSprite (const char* /*mesh*/, const char* /*skin*/,
  float /*x*/, float /*y*/, float /*z*/, float /*scale*/, bool /*lit*/,
  float /*alpha*/)
{
/*
  I didn't change this function to use the new converter since it seems
  to be unused
  - mgeisse

  printf ("Adding sprite %s ...\n", mesh);

  csCrossBuild_SpriteTemplateFactory builder (object_reg);
  csMatrix3 m;

  LevelLoader->LoadTexture (skin, skin);
  converter* filedata = new converter;
  filedata->ivcon (mesh, true, false, NULL, myVFS);
  iMeshObjectFactory * t = (iMeshObjectFactory*)builder.CrossBuild(*filedata);
  iSprite3DFactoryState* fstate = SCF_QUERY_INTERFACE (t, iSprite3DFactoryState);
  fstate->SetMaterialWrapper (engine->GetMaterialList ()
  	->FindByName (skin));
  iMeshFactoryWrapper* meshwrap = engine->CreateMeshFactory (NULL);
  meshwrap->SetMeshObjectFactory (t);

  iSpriteAction * action = fstate->AddAction ();
  int i;
  for (i=0; i < fstate->GetFrameCount (); i++)
    action->AddFrame (fstate->GetFrame (i), 100);
  action->SetName ("all");

  iMeshWrapper* s = engine->CreateMeshWrapper (SCF_QUERY_INTERFACE (meshwrap,
  	iMeshFactoryWrapper), "", SCF_QUERY_INTERFACE (start, iSector),
	csVector3 (x, y, z));
  iSprite3DState* state = SCF_QUERY_INTERFACE (s->GetMeshObject (), iSprite3DState);
  state->SetAction ("all");      state->SetFrame (0);
  m.Identity();  m=m*scale;  s->GetMovable ()->SetTransform (m);
  s->GetMovable ()->UpdateMove ();
  if (lit) s->DeferUpdateLighting (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);
  if (alpha < 1 && alpha > 0) state->SetMixMode (CS_FX_SETALPHA (alpha));
  if (alpha == 2) state->SetMixMode (CS_FX_KEYCOLOR);
  meshobjs.Push (s);
  fstate->DecRef ();
  state->DecRef ();

  delete filedata;
*/
}

void sqkSystem::CreateViews ()
{
  printf ("Creating views ...\n");

  while (views.Length())
    views.Delete (views.Length() - 1);

  sqkView * v;
  const char * x_str, * y_str, * x_ptr, * y_ptr, * center;
  const char * xm_str, * ym_str, * xm_ptr, * ym_ptr;
  float x, y, xmargin, ymargin;
  bool do_margin = false;

  char section [32];
  sprintf (section, "Squawk.ViewLayout(%d).View(1)", viewLayout);


  // Look for the start sector in this map.
  iCameraPosition *cp = engine->GetCameraPositions ()->FindByName ("Start");
  const char *room_name;
  if (cp)
    room_name = cp->GetSector ();
  else
    room_name = "room";

  start = engine->GetSectors ()->FindByName (room_name);
  if (!start)
    message_cleanup_and_exit ("Can't find starting room!\n");

  iConfigFile* cfg = CS_QUERY_REGISTRY (object_reg, iConfigManager);

  int i;
  for (i = 0; i < 9999; i++)
  {
    printf ("Loading view [%s]\n", section);

    v = new sqkView ();

    csString Keyname;
    Keyname.Clear() << section << ".X";
    x_ptr  = x_str  = cfg->GetStr (Keyname, "0,1,1,0");
    Keyname.Clear() << section << ".Y";
    y_ptr  = y_str  = cfg->GetStr (Keyname, "1,1,0,0");
    Keyname.Clear() << section << ".XMargin";
    xm_ptr = xm_str = cfg->GetStr (Keyname, "!");
    Keyname.Clear() << section << ".YMargin";
    ym_ptr = ym_str = cfg->GetStr (Keyname, "!");
    if ((ym_ptr[0] != '!') && (ym_ptr[0] != '!'))
      do_margin = true;

    v->AddView (new csView (engine, myG3D));
    int current_view = 0;
    while (x_ptr [0])
    {
      x = next_float (&x_ptr) * FrameWidth;
      y = next_float (&y_ptr) * FrameHeight;

      if (do_margin)
      {
        // I multiply both margins by screen height so the margin will be
        // the same thickness horizontally and vertically.
        xmargin = viewMargin * next_float (&xm_ptr) * FrameHeight;
        ymargin = viewMargin * next_float (&ym_ptr) * FrameHeight;

        // shave 1 pixel from the bottom and right sides of the view
        if (xmargin < 0) xmargin--;
        if (ymargin < 0) ymargin--;

        x += xmargin;
        y += ymargin;
      }

      if (x > FrameWidth)  x = FrameWidth;
      if (y > FrameHeight) y = FrameHeight;
      if (x < 0) x = 0;
      if (y < 0) y = 0;

      v->AddVertex (current_view, (int)x, (int)y);

      if (x_ptr[0] == ':')
      {
        v->AddView (new csView (engine, myG3D));
        current_view++; x_ptr++; y_ptr++;
        if (do_margin) {xm_ptr++; ym_ptr++;}
      }
    }

    Keyname.Clear() << section << ".FOV";
    v->SetFOV   (cfg->GetFloat (Keyname, 1.0f) * FrameHeight, FrameWidth);
    Keyname.Clear() << section << ".Gravity";
    v->gravity = cfg->GetBool (Keyname, true);
    Keyname.Clear() << section << ".Center";
    center     = cfg->GetStr   (Keyname, "0.5,0.5");
    csScanStr (center, "%f,%f", &x, &y);
    v->SetCenter (x * FrameWidth, y * FrameHeight);

    v->SetCamera (cp, engine);

    views.Push (v);

    sprintf (section, "Squawk.ViewLayout(%d).View(%d)", viewLayout, i + 2);
    if (!(cfg->SubsectionExists (section))) break;
  }
  cfg->DecRef ();
}

static bool SqEventHandler (iEvent& ev)
{
  if (ev.Type == csevBroadcast && ev.Command.Code == cscmdProcess)
  {
    Sys->SetupFrame ();
    return true;
  }
  else if (ev.Type == csevBroadcast && ev.Command.Code == cscmdFinalProcess)
  {
    Sys->FinishFrame ();
    return true;
  }
  else if (ev.Type == csevBroadcast && ev.Command.Code == cscmdCommandLineHelp)
  {
    Sys->Help ();
    return true;
  }
  else
  {
    return Sys->SqHandleEvent (ev);
  }
}

void sqkSystem::InitApp (int argc, char* argv[], const char* iConfigName)
{
#ifdef CS_DEBUG
  // enable all kinds of useful FPU exceptions on a x86
  _control87 (0x33, 0x3f);
#else
  // this will disable exceptions on DJGPP (for the "industrial" version)
  _control87 (0x3f, 0x3f);
#endif

  object_reg = csInitializer::CreateEnvironment ();
  if (!object_reg)
    message_cleanup_and_exit ("bad init!\n");

  if (!csInitializer::SetupConfigManager (object_reg, iConfigName))
    message_cleanup_and_exit ("bad init!\n");

  csInitializer::SetupCommandLineParser (object_reg, argc, argv);
  if (!csInitializer::RequestPlugins (object_reg,
  	CS_REQUEST_REPORTER,
  	CS_REQUEST_REPORTERLISTENER,
  	CS_REQUEST_END
	))
    message_cleanup_and_exit ("bad init!\n");

  SetDefaults ();

  if (!csInitializer::SetupEventHandler (object_reg, SqEventHandler))
    message_cleanup_and_exit ("bad init!\n");

  // Check for commandline help.
  if (csCommandLineHelper::CheckHelp (object_reg))
  {
    csCommandLineHelper::Help (object_reg);
    exit (0);
  }

  vc = CS_QUERY_REGISTRY (object_reg, iVirtualClock);

  kbd = CS_QUERY_REGISTRY (object_reg, iKeyboardDriver);
  if (!kbd)
    message_cleanup_and_exit ("no generic keyboard driver!\n");

  LevelLoader = CS_QUERY_REGISTRY (object_reg, iLoader);
  if (!LevelLoader)
    message_cleanup_and_exit ("no level loader plug-in!\n");

  myG3D = CS_QUERY_REGISTRY (object_reg, iGraphics3D);
  if (!myG3D)
    message_cleanup_and_exit ("no graphics3d plug-in!\n");

  myG2D = CS_QUERY_REGISTRY (object_reg, iGraphics2D);
  if (!myG2D)
    message_cleanup_and_exit ("no graphics2d plug-in!\n");

  myVFS = CS_QUERY_REGISTRY (object_reg, iVFS);
  if (!myVFS)
    message_cleanup_and_exit ("no vfs plug-in!\n");

  if (!csInitializer::OpenApplication (object_reg))
    message_cleanup_and_exit ("Error opening system!\n");

  FrameWidth = myG2D->GetWidth ();
  FrameHeight = myG2D->GetHeight ();

  iTextureManager* txtmgr = myG3D->GetTextureManager ();
  txtmgr->PrepareTextures ();
  txtmgr->SetPalette ();

  printf ("%s\n", SQK_VERSION);

  // Find the pointer to engine plugin
  engine = CS_QUERY_REGISTRY (object_reg, iEngine);
  if (!engine)
  {
    printf ("No iEngine plugin!\n");
    exit (-1);
  }

  engine->SetLightingCacheMode (0);

  // load map

  iConfigManager* cfg = CS_QUERY_REGISTRY (object_reg, iConfigManager);
  if (mapfile == NULL)
  {
    const char * wf = cfg->GetStr ("Squawk.World", "sqk1");
    mapfile = new char [strlen (wf) + 6];
    sprintf (mapfile, "/lev/%s", wf);
  }
  printf ("Loading map '%s'...\n", mapfile);
  if (!myVFS->ChDir (mapfile))
    message_cleanup_and_exit ("The map file does not exist!\n");
  if (!LevelLoader->LoadMapFile ("world"))
    message_cleanup_and_exit ("Loading of map failed!\n");

  // @@@ UGLY HACK: Change to other directory before doing Prepare()
  // because otherwise precalc_info file will be written into sqk1.
  // The root myVFS directory is usually unwritable, which is what we need.
  myVFS->ChDir ("/");

  engine->Prepare ();
  txtmgr->SetPalette ();

  viewMargin = cfg->GetFloat ("Squawk.Margin", viewMargin);
  viewLayout = cfg->GetInt   ("Squawk.Layout", viewLayout);
  cfg->DecRef ();
  CreateViews ();

  iConsoleOutput *pConsole = CS_QUERY_REGISTRY (object_reg, iConsoleOutput);
  if (pConsole)
  {
    pConsole->AutoUpdate (false);
    pConsole->DecRef ();
  }

  if (myG2D->BeginDraw())
  {
    myG2D->ClearAll (0);
    myG2D->FinishDraw ();
  }
}

void sqkSystem::eatkeypress (int key, bool shift, bool alt, bool ctrl)
{
  (void)shift; (void)alt; (void)ctrl;

  switch (key)
  {
    case CSKEY_ESC :
      {
        iEventQueue* q = CS_QUERY_REGISTRY (object_reg, iEventQueue);
        if (q)
	{
	  q->GetEventOutlet()->Broadcast (cscmdQuit);
	  q->DecRef ();
        }
      }
      break;

    case '1' : viewLayout = 1; CreateViews(); break;
    case '2' : viewLayout = 2; CreateViews(); break;

    case '3' : viewMargin = 0.00; CreateViews(); break;
    case '4' : viewMargin = 0.01; CreateViews(); break;
    case '5' : viewMargin = 0.02; CreateViews(); break;
  }
}

void sqkSystem::SetupFrame ()
{
  csTicks elapsed_time, current_time;
  elapsed_time = vc->GetElapsedTicks ();
  current_time = vc->GetCurrentTicks ();

  int i;

  // Animate meshes

  if (frameTime == 0)  frameTime = current_time + 50;
  if (current_time > frameTime)
  {
    frameTime = current_time + 50;
    for (i=0; i < meshobjs.Length(); i++)
    {
      iMeshWrapper* sp = GetMeshWrapper (i);
      iSprite3DState* state = SCF_QUERY_INTERFACE (sp->GetMeshObject (), iSprite3DState);
      int frame = state->GetCurFrame();
      if (frame >= state->GetFrameCount()-1) state->SetFrame(0);
      else state->SetFrame(frame + 1);
      state->DecRef ();
      //GetSprite(i)->DeferUpdateLighting (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);
    }
  }

  // Update mesh lighting

  for (i=0; i < engine->GetMeshes ()->GetCount (); i++)
    engine->GetMeshes ()->Get (i)->DeferUpdateLighting
      (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);

  // Draw the views

  if (!myG3D->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS)) return;
  for (i=0; i < views.Length(); i++)
    GetView(i)->Draw ();

  sqkView* view;
  for (i=0; (view = GetView (i)); i++)
  {
    float yaw = 0, pitch = 0, roll = 0, move = 0;
    switch (i)
    {
      case 0:
        if (kbd->GetKeyState (CSKEY_RIGHT)) ++yaw;
        if (kbd->GetKeyState (CSKEY_LEFT) ) --yaw;
        if (kbd->GetKeyState (CSKEY_UP)   ) --pitch;
        if (kbd->GetKeyState (CSKEY_DOWN) ) ++pitch;
        if (kbd->GetKeyState (CSKEY_PGUP) ) ++move;
        if (kbd->GetKeyState (CSKEY_PGDN) ) --move;   
        if (kbd->GetKeyState (CSKEY_INS)  ) ++roll;
        if (kbd->GetKeyState (CSKEY_DEL)  ) --roll;
        break;

      case 1:
        if (kbd->GetKeyState ('d')) ++yaw;
        if (kbd->GetKeyState ('a')) --yaw;
        if (kbd->GetKeyState ('w')) --pitch;
        if (kbd->GetKeyState ('s')) ++pitch;
        if (kbd->GetKeyState ('t')) ++move;
        if (kbd->GetKeyState ('g')) --move;   
        if (kbd->GetKeyState ('r')) ++roll;
        if (kbd->GetKeyState ('f')) --roll;
        break;

      case 2:
        if (kbd->GetKeyState ('l')) ++yaw;
        if (kbd->GetKeyState ('j')) --yaw;
        if (kbd->GetKeyState ('i')) --pitch;
        if (kbd->GetKeyState ('k')) ++pitch;
        if (kbd->GetKeyState ('t')) ++move;
        if (kbd->GetKeyState ('g')) --move;   
        if (kbd->GetKeyState ('y')) ++roll;
        if (kbd->GetKeyState ('h')) --roll;
        break;
    }
    if (yaw || pitch || roll) view->Rotate (csVector3 (pitch, yaw, roll), 0.001 * elapsed_time);
    if (move) view->Move (VEC_FORWARD, 0.001 * elapsed_time * move);
  }
}

void sqkSystem::FinishFrame ()
{
  myG3D->FinishDraw ();
  myG3D->Print (NULL);
}

bool sqkSystem::SqHandleEvent (iEvent &Event)
{
  switch (Event.Type)
  {
    case csevKeyDown:
      eatkeypress (Event.Key.Code,
        !!(Event.Key.Modifiers & CSMASK_SHIFT),
        !!(Event.Key.Modifiers & CSMASK_ALT),
        !!(Event.Key.Modifiers & CSMASK_CTRL));
      return true;
  }
  return false;
}

void sqkSystem::Help()
{
  printf("  -cfg <file>        "
    "alternate configuration file (default=/config/squawk.cfg)\n");
}

void sqkSystem::SetDefaults ()
{
  iCommandLineParser* cmdline = CS_QUERY_REGISTRY (object_reg,
  	iCommandLineParser);

  const char *val;
  if ((val = cmdline->GetName ()))
  {
    mapfile = new char [strlen (val) + 6];
    sprintf (mapfile, "/lev/%s", val);
  }

  if ((val = cmdline->GetOption ("layout")))
    viewLayout = atol (val);

  if ((val = cmdline->GetOption ("margin")))
    csScanStr (val, "%f", &viewMargin);
  cmdline->DecRef ();
}

