/*
    Copyright (C) 1999 by Seth Galbraith

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SQKSYSTM_H
#define SQUSYSTM_H

#include "iengine/mesh.h"
#include "csgeom/math3d.h"
#include "csutil/csvector.h"
#include "sqkview.h"

struct iEngine;
struct iLoader;
struct iKeyboardDriver;
struct iVirtualClock;
struct iEvent;
struct iGraphics3D;
struct iGraphics2D;
struct iVFS;

CS_DECLARE_TYPED_IBASE_VECTOR (csMeshWrapVector, iMeshWrapper);
CS_DECLARE_TYPED_VECTOR (sqkViewVector, sqkView);

class sqkSystem
{
  public:
    int FrameWidth, FrameHeight;
    iEngine* engine;
    sqkViewVector views;
    csMeshWrapVector meshobjs;
    csTicks frameTime;
    char * gamecfg;
    char * mapfile;
    iSector * start;
    int viewLayout;
    float viewMargin;
    iLoader *LevelLoader;
    iGraphics3D *myG3D;
    iGraphics2D *myG2D;
    iVFS *myVFS;
    iKeyboardDriver* kbd;
    iVirtualClock* vc;
    iObjectRegistry* object_reg;

  public:

    sqkSystem ();
    virtual ~sqkSystem ();

    void InitApp (int argc, char* argv[], const char* cfg);
    void Help();
    void SetDefaults ();
    void SetupFrame ();
    void FinishFrame ();
    bool SqHandleEvent (iEvent &Event);
    void eatkeypress (int key, bool shift, bool alt, bool ctrl);

    sqkView* GetView (int v)
      {return (v < views.Length ()) ? (sqkView*)views [v] : (sqkView*)NULL;}
    iMeshWrapper* GetMeshWrapper (int s)
      {return (s < meshobjs.Length ()) ? meshobjs [s] : (iMeshWrapper*)NULL;}

    void CreateLights ();
    void CreateSkyBox ();
    void CreateSprites ();
    void CreateViews ();
    void AddSprite (const char* mesh, const char* skin, float x, float y, float z,
      float scale, bool lit, float alpha);
};

#endif /*SQKSYSTM_H*/
