/*
    Copyright (C) 1999 by Seth Galbraith

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "sqkview.h"

#include "csgeom/transfrm.h"
#include "iengine/engine.h"
#include "iengine/campos.h"
#include "iengine/camera.h"

sqkView::sqkView()
{
  gravity = true;
}

sqkView::~sqkView()
{
}

void sqkView::AddView (iView * v)
{
  views.Push (v);
}

void sqkView::AddVertex (int v, int x, int y)
{
  GetView(v)->AddViewVertex (x, y);
}

void sqkView::Rotate (const csVector3& v, float speed)
{
  int i;
  for (i = 0; i < views.Length(); i++)
  {
    iCamera* c = GetView(i)->GetCamera();
    csOrthoTransform& camtrans = c->GetTransform ();
    if (gravity)
    {
      if (v.x || v.z) camtrans.RotateThis (csVector3 (v.x,0,v.z), speed);
      if (v.y)
      {
        if (camtrans.GetO2T().Col2().y < 0) // upside down (-:
          camtrans.RotateOther (csVector3 (0,-(v.y),0), speed);
        else
	  camtrans.RotateOther (csVector3 (0,v.y,0), speed);
      }
    }
    else
      camtrans.RotateThis (v,speed);
  }
}

void sqkView::Move (const csVector3& v, float speed)
{
  csVector3 u = v * speed;
  int i;
  for (i = 0; i < views.Length(); i++)
    GetView(i)->GetCamera()->Move (u, false);
}

void sqkView::SetCenter (float x, float y)
{
  int i;
  for (i=0; i < views.Length(); i++)
    GetView(i)->GetCamera ()->SetPerspectiveCenter (x, y);
}

void sqkView::SetSector (iSector * s)
{
  int i;
  for (i=0; i < views.Length(); i++)
    GetView(i)->GetCamera ()->SetSector (s);
}

void sqkView::SetPosition (const csVector3 & v)
{
  int i;
  for (i=0; i < views.Length(); i++)
    GetView(i)->GetCamera ()->GetTransform ().SetOrigin (v);
}

void sqkView::SetFOV (float f, int w)
{
  int i;
  for (i=0; i < views.Length(); i++)
    GetView(i)->GetCamera ()->SetFOV ((int)f, w);
}

void sqkView::Draw ()
{
  int i;
  for (i=0; i < views.Length(); i++)
    GetView(i)->Draw();
}

void sqkView::SetCamera (iCameraPosition *cp, iEngine *engine)
{
  int i;
  for (i=0; i < views.Length(); i++)
    cp->Load (GetView(i)->GetCamera (), engine);
}
