/*
    Copyright (C) 1999 by Seth Galbraith

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SQKVIEW_H
#define SQKVIEW_H

#include "csutil/typedvec.h"
#include "csgeom/vector3.h"
#include "ivaria/view.h"

struct iCameraPosition;
struct iSector;

CS_DECLARE_TYPED_IBASE_VECTOR (csViewVector, iView);

/**
 * A viewport that doesn't have to be convex or even contiguous,
 * made up of multiple convex views
 */
class sqkView
{
  public:

    csViewVector views;

    /**
     * If true (it is by default), this view rotates intuitively.
     * Left and right rotation is always around the world's Y axis.
     * @@@ this should be replaced by a gravity axis or something.
     * Left and right are reversed when the camera is upside down.
     * If false, left and right rotation is around the view's Y axis.
     */
    bool gravity;

  public:

    sqkView();
    ~sqkView();

    /// Add a convex view.
    void AddView (iView * v);

    /// Get a convex view by index.
    iView* GetView (int v)
      {return (v < views.Length ()) ? views [v] : (iView*)NULL;}

    /// Add a vertex to a view clipping polygon
    void AddVertex (int v, int x, int y);

    /// Set the center of the view on the screen
    void SetCenter (float x, float y);

    /// Set the depth of the feild-of-view
    void SetFOV (float f, int w);

    /**
     * Relative rotation, measured in radians. (180 degrees = pi * radians)
     * v.x = pitch = looking up and down
     * v.y = yaw   = turning left and right
     * v.z = roll  = leaning left and right
     */
    void Rotate (const csVector3& v, float speed);
    void Move ( const csVector3& v, float speed);

    void SetSector (iSector * s);
    void SetPosition (const csVector3& v);
    void SetCamera (iCameraPosition *cp, iEngine *engine);
    void Draw ();
};

#endif /*SQKVIEW_H*/
