/*
    Copyright (C) 1999 by Seth Galbraith

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "cssys/sysfunc.h"
#include "squawk.h"
#include "sqkview.h"
#include "sqksystm.h"

#include "csutil/scanstr.h"
#include "cstool/initapp.h"

#include "iutil/vfs.h"
#include "ivideo/graph2d.h"
#include "ivideo/graph3d.h"
#include "ivideo/txtmgr.h"

CS_IMPLEMENT_APPLICATION

sqkSystem* Sys = NULL;

void Cleanup ()
{
  csPrintf ("Cleaning up...\n");
  iObjectRegistry* object_reg = Sys->object_reg;
  delete Sys; Sys = NULL;
  csInitializer::DestroyApplication (object_reg);
}

void message_cleanup_and_exit (const char * s)
{
  csPrintf ("%s", s);
  Cleanup ();
  exit (1);
}

int main (int argc, char* argv[])
{
  const char* cfg = "/config/squawk.cfg";

  int i;
  for (i=0; i < argc-1; i++)
   if (strcmp (argv[i], "-cfg") == 0)
     {cfg = argv[i+1]; break;}

  Sys = new sqkSystem ();
  Sys->InitApp (argc, argv, cfg);
  csDefaultRunLoop(Sys->object_reg);
  Cleanup ();
  return 0;
}

float next_float (const char ** ptr)
{
  // find the next float in a list of floats seperated by commas and whitespace,
  // and return it's value, leaving the string pointer at the end of the float
  float f;
  *ptr += strspn (*ptr, ", \t");
  csScanStr (*ptr, "%f", &f);
  *ptr += strspn (*ptr, "0123456789.eE+- \t");
  return f;
}
