/*
    Copyright (C) 2000 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __CS_PS2D_H__
#define __CS_PS2D_H__

#include "video/canvas/common/graph2d.h"
#include "iutil/event.h"

class csGraphics2Dps2 : public csGraphics2D, public iEventPlug
{
  iEventOutlet *EventOutlet;

public:
  SCF_DECLARE_IBASE_EXT(csGraphics2D);

  csGraphics2Dps2 (iBase*);
  virtual ~csGraphics2Dps2 ();

  virtual bool Initialize (iObjectRegistry*);
  virtual bool Open ();
  virtual void Close ();

  virtual bool BeginDraw ();
  virtual void FinishDraw ();
  virtual void Print (csRect *area = NULL);
  virtual void SetRGB (int i, int r, int g, int b);

  virtual void DrawLine (float x1, float y1, float x2, float y2, int color);
  virtual void Clear (int color);
  virtual void Write (int x, int y, int fg, int bg, const char *text);

  /// Set mouse position.
  virtual bool SetMousePosition (int x, int y);

  /// Set mouse cursor shape
  virtual bool SetMouseCursor (csMouseCursorID iShape);

  /// Called on every frame by system driver
  virtual bool HandleEvent (iEvent &Event);

  virtual unsigned char *GetPixelAt (int x, int y);

  //------------------------ iEventPlug interface ---------------------------//
  virtual unsigned GetPotentiallyConflictingEvents ()
  { return CSEVTYPE_Keyboard | CSEVTYPE_Mouse; }
  virtual unsigned QueryEventPriority (unsigned /*iType*/)
  { return 150; }
};

#endif // __CS_PS2D_H__
